import os
from tempfile import TemporaryDirectory

from nvbenjo import benchmark, cfg
from nvbenjo.utils import PrecisionType


def test_pytorch_simple():
    with TemporaryDirectory() as tmpdir:
        num_batches = 2
        model_cfg = cfg.TorchModelConfig(
            name="torch-shufflenet-v2-x0-5",
            type_or_path="torchvision:shufflenet_v2_x0_5",
            shape=(("B", 3, 224, 224),),
            devices=["cpu"],
            batch_sizes=[1],
            num_warmup_batches=1,
            num_batches=num_batches,
            runtime_options={
                "test1": cfg.TorchRuntimeConfig(
                    compile=False,
                    precision=PrecisionType.FP32,
                    enable_profiling=True,
                    profiling_prefix=os.path.join(tmpdir, "profile_"),
                    profiler_kwargs={"profile_memory": True, "record_shapes": True},
                ),
            },
            custom_batchmetrics={
                "fps": 1.0,
            },
        )
        results = benchmark.benchmark_models({"model_1": model_cfg})
        assert not results.empty
        assert "model_1" in results.model.to_numpy()
        assert "test1" in results.runtime_options.to_numpy()
        assert len(results.time_inference.to_numpy()) == num_batches

        # Check that profiling files were created
        profile_files = os.listdir(tmpdir)
        assert len(profile_files) == 1
        assert profile_files[0].startswith("profile_")
        assert profile_files[0].endswith(".json")



import pytest
import torch

from nvbenjo.utils import EXAMPLE_VALID_SHAPES, format_num, format_seconds, get_rnd_from_shape_s, NoBatchShapeError


def test_format_seconds():
    assert format_seconds(1.001) == "1.001 s"
    assert format_seconds(0.001001) == "1.001 ms"
    assert format_seconds(0.000001001) == "1.001 us"


def test_format_num():
    assert format_num(1000) == "1.00 K"
    assert format_num(1000000) == "1.00 M"
    assert format_num(1000000000) == "1.00 G"
    assert format_num(1000000000000) == "1.00 T"
    assert format_num(1000000000000000) == "1.00 P"
    assert format_num(1024, bytes=True) == "1.00 KB"
    assert format_num(1048576, bytes=True) == "1.00 MB"
    assert format_num(1073741824, bytes=True) == "1.00 GB"
    assert format_num(1099511627776, bytes=True) == "1.00 TB"
    assert format_num(1125899906842624, bytes=True) == "1.00 PB"


def test_get_rnd_shape_examples():
    for example_shape in EXAMPLE_VALID_SHAPES:
        _ = get_rnd_from_shape_s(example_shape, batch_size=12)
    _ = get_rnd_from_shape_s(({"name": "input1", "type": "float", "shape": ("B", 6), "value": 1},), batch_size=12)
    with pytest.raises(NoBatchShapeError, match="Please ensure that the shape contains an identifier"):
        _ = get_rnd_from_shape_s(
            ({"name": "input1", "type": "float", "shape": (6,), "value": [1, 2, 3, 4, 5, 6]},), batch_size=12
        )
    _ = get_rnd_from_shape_s(
        (
            {"name": "input1", "type": "float", "shape": (6,), "value": [1, 2, 3, 4, 5, 6]},
            {"name": "input1", "type": "float", "shape": (6, "B"), "value": 1},
        ),
        batch_size=12,
    )


def test_get_rnd_shape_invalid():
    example_shape = ("B", 3, 224, 224)
    with pytest.raises(ValueError):
        _ = get_rnd_from_shape_s(example_shape, batch_size=12, min_val=0, max_val=1, dtype="invalid_dtype")
    with pytest.raises(ValueError):
        _ = get_rnd_from_shape_s(example_shape, batch_size=12, min_val="invalid", max_val=1)
    with pytest.raises(ValueError):
        _ = get_rnd_from_shape_s(example_shape, batch_size=12, min_val=0, max_val="invalid")
    with pytest.raises(ValueError):
        _ = get_rnd_from_shape_s((1, 3, 224, 224), batch_size=12)  # missing batch size identifier

    with pytest.raises(ValueError):
        _ = get_rnd_from_shape_s(tuple(), batch_size=12, min_val=0, max_val=1)


def test_get_rnd_shape_valid():
    example_shape = ("B", 3, 224, 224)
    rnd_tensor, set_individual_dtype = get_rnd_from_shape_s(
        example_shape, batch_size=12, min_val=0, max_val=1, dtype="float32"
    )
    assert torch.all(rnd_tensor >= 0) and torch.all(rnd_tensor <= 1)
    assert not set_individual_dtype
    assert rnd_tensor.shape == (12, 3, 224, 224)
    assert rnd_tensor.dtype == torch.float32

    rnd_tensor, _ = get_rnd_from_shape_s(example_shape, batch_size=12, min_val=0, max_val=1, dtype="int32")
    assert rnd_tensor.shape == (12, 3, 224, 224)
    assert rnd_tensor.dtype == torch.int32

    example_shape = (("B", 3, 224, 224), (1, 3, 12, 13))
    rnd_tensor, _ = get_rnd_from_shape_s(example_shape, batch_size=12, min_val=0, max_val=1, dtype="float32")
    assert rnd_tensor[0].shape == (12, 3, 224, 224)
    assert rnd_tensor[1].shape == (1, 3, 12, 13)

    example_shape = (
        {"name": "input1", "type": "float", "shape": ("B", 3, 224, 224), "min_max": (3, 12)},
        {"name": "input2", "type": "int", "shape": (1, 3), "min_max": (0, 1)},
    )
    rnd_tensor, set_individual_dtype = get_rnd_from_shape_s(example_shape, batch_size=12, dtype="float32")
    assert set_individual_dtype
    assert rnd_tensor["input1"].shape == (12, 3, 224, 224)
    assert rnd_tensor["input1"].dtype == torch.float32
    assert torch.all(rnd_tensor["input1"] >= 3) and torch.all(rnd_tensor["input1"] <= 12)
    assert rnd_tensor["input2"].shape == (1, 3)
    assert rnd_tensor["input2"].dtype == torch.int32
    assert torch.all(rnd_tensor["input2"] >= 0) and torch.all(rnd_tensor["input2"] <= 1)



import itertools
import logging
import time
import typing as ty
from typing import Any, Callable, Optional, Dict

import numpy as np
import pandas as pd
import torch
import torch.nn as nn
from rich import progress
from rich.progress import Progress

import nvbenjo.utils as utils
from nvbenjo import console, torch_utils
from nvbenjo.cfg import BaseModelConfig, TorchModelConfig, OnnxModelConfig, TorchRuntimeConfig, OnnxRuntimeConfig

logger = logging.getLogger(__name__)


def load_model(
    type_or_path: str, device: torch.device, runtime_config: TorchRuntimeConfig | OnnxRuntimeConfig, **kwargs
) -> Any:
    """Load a model, may be a PyTorch or ONNX model based on the runtime configuration.

    Parameters
    ----------
    type_or_path : str
        String specifying the model type or path
    device : torch.device
        Device to load the model onto
    runtime_config : :class:`~nvbenjo.cfg.TorchRuntimeConfig` or :class:`~nvbenjo.cfg.OnnxRuntimeConfig`
        Runtime configuration for the model

    Returns
    -------
    Any
        Loaded model instance
    """
    match runtime_config:
        case OnnxRuntimeConfig():
            from nvbenjo import onnx_utils

            return onnx_utils.get_model(type_or_path, device=device, runtime_config=runtime_config, **kwargs)
        case TorchRuntimeConfig():
            return torch_utils.get_model(type_or_path, device=device, runtime_config=runtime_config, **kwargs)
        case _:
            raise ValueError(f"Unknown runtime config type {type(runtime_config)}")


def _test_load_models(model_cfgs: Dict[str, BaseModelConfig]) -> None:
    loaded_types = []
    logger.info("Checking if models are valid and available")
    for _, model_cfg in model_cfgs.items():
        if model_cfg.type_or_path not in loaded_types:
            initial_runtime_options = list(model_cfg.runtime_options.values())[0]
            _ = load_model(
                model_cfg.type_or_path,
                device=torch.device("cpu"),
                runtime_config=initial_runtime_options,
                **model_cfg.kwargs,
            )
            loaded_types.append(model_cfg.type_or_path)


def benchmark_models(model_cfgs: Dict[str, BaseModelConfig], measure_memory: Optional[bool] = True) -> pd.DataFrame:
    """Benchmark the given models.

    Parameters
    ----------
    model_cfgs : Dict[str, :class:`~nvbenjo.cfg.TorchModelConfig` | :class:`~nvbenjo.cfg.OnnxModelConfig`]
    measure_memory : bool, optional
        Whether to measure memory usage during benchmarking, by default True

    Returns
    -------
    pd.DataFrame
        A DataFrame containing the benchmarking results


    Examples
    --------
    Basic usage with single PyTorch model::

        from nvbenjo import cfg
        from nvbenjo.utils import PrecisionType
        from nvbenjo import benchmark

        model_cfg = cfg.TorchModelConfig(
            name="torch-shufflenet-v2-x0-5",
            type_or_path="torchvision:shufflenet_v2_x0_5",
            shape=(("B", 3, 224, 224),),
            devices=["cpu"],
            batch_sizes=[1],
            num_warmup_batches=1,
            num_batches=2,
            runtime_options={
                "test1": cfg.TorchRuntimeConfig(compile=False, precision=PrecisionType.FP32),
            },
            custom_batchmetrics={
                "fps": 1.0,
            },
        )
        results = benchmark.benchmark_models({"model_1": model_cfg})
    """
    _test_load_models(model_cfgs)

    with _get_progress_bar() as progress_bar:
        model_task = progress_bar.add_task("Benchmarking models", total=len(model_cfgs))
        results = []

        for model_name, model_cfg in model_cfgs.items():
            progress_bar.update(model_task, description=f"Benchmarking {model_name}")
            model_results = benchmark_model(model_cfg, progress_bar=progress_bar, measure_memory=measure_memory)
            model_results["model"] = model_name
            if model_cfg.custom_batchmetrics:
                model_results = utils.calculate_batchmetrics(model_results, model_cfg.custom_batchmetrics)
            results.append(model_results)
            progress_bar.advance(model_task)

        results = pd.concat(results)

    return results


def _get_progress_bar() -> Progress:
    return Progress(
        "[progress.description]{task.description}",
        progress.BarColumn(bar_width=80),
        "[progress.percentage]{task.completed}/{task.total}",
        console=console,
    )


def _run_warmup(
    model: nn.Module,
    batch: utils.TensorLike,
    device: torch.device,
    num_warmup_batches: int,
    progress_bar: Optional[Progress],
):
    try:
        if progress_bar is not None:
            warm_up_task = progress_bar.add_task("    Warm-up", total=num_warmup_batches)

        for _ in range(num_warmup_batches):
            batch = torch_utils.transfer_to_device(batch, device)
            r = torch_utils.run_model_with_input(model, batch)
            try:
                _ = torch_utils.transfer_to_device(r, to_device=torch.device("cpu"))
            except Exception:
                console.print(utils.TRANSFER_WARNING)

            if progress_bar is not None:
                progress_bar.advance(warm_up_task)

    finally:
        if progress_bar is not None:
            progress_bar.remove_task(warm_up_task)


def _measure_timings(
    model: Any,
    batch: utils.TensorLike,
    batch_size: int,
    device: torch.device,
    num_batches: int,
    progress_bar: Optional[Progress],
    timing_function: Callable = torch_utils.measure_repeated_inference_timing,
    profiler: Optional[torch.profiler.profile] = None,
) -> pd.DataFrame:
    if progress_bar is not None:
        measure_task = progress_bar.add_task(
            "    Inference",
            total=num_batches,
        )

    def progress_callback():
        if progress_bar is not None:
            progress_bar.advance(measure_task)
        else:
            pass
        if profiler is not None:
            profiler.step()

    try:
        cur_raw_results = timing_function(
            model,
            batch,
            batch_size,
            device,
            num_runs=num_batches,
            progress_callback=progress_callback,
        )
    finally:
        if progress_bar is not None:
            progress_bar.remove_task(measure_task)
    return cur_raw_results


def _get_device(runtime_config: OnnxRuntimeConfig | TorchRuntimeConfig, device: str, console) -> torch.device:
    device_chosen = torch.device(device)
    match device_chosen.type:
        case "cpu":
            match runtime_config:
                case TorchRuntimeConfig():
                    return device_chosen
                case OnnxRuntimeConfig():
                    # If CUDA is available and the execution provider is set to CUDA or TensorRT, use CUDA
                    from nvbenjo import onnx_utils

                    available_providers = onnx_utils.ort.get_available_providers()  # type: ignore
                    if (
                        torch.cuda.is_available()
                        and (
                            "CUDAExecutionProvider" in available_providers
                            or "TensorrtExecutionProvider" in available_providers
                        )
                        # Check for
                        # [("CUDAExecutionProvider", {"device_id": 0})] style entries
                        and runtime_config.execution_providers
                        and any(
                            provider in ["CUDAExecutionProvider", "TensorrtExecutionProvider"]
                            or provider is not None
                            and provider[0] in ["CUDAExecutionProvider", "TensorrtExecutionProvider"]
                            for provider in runtime_config.execution_providers  # type: ignore
                        )
                    ):
                        device_id = 0
                        for provider in runtime_config.execution_providers:  # type: ignore
                            if len(provider) == 2 and provider[0] in [
                                "CUDAExecutionProvider",
                                "TensorrtExecutionProvider",
                            ]:
                                device_id = provider[1].get("device_id", 0)  # type: ignore
                        device_chosen = torch.device(f"cuda:{device_id}")
            return device_chosen
        case "cuda":
            match runtime_config:
                case TorchRuntimeConfig():
                    if torch.cuda.is_available():
                        return device_chosen
                    else:
                        console.print("[yellow]CUDA is not available. Running on CPU.[/yellow]")
                        return torch.device("cpu")

                case OnnxRuntimeConfig():
                    from nvbenjo import onnx_utils

                    available_providers = onnx_utils.ort.get_available_providers()  # type: ignore
                    if torch.cuda.is_available() and (
                        "CUDAExecutionProvider" in available_providers
                        or "TensorrtExecutionProvider" in available_providers
                    ):
                        return device_chosen
                    else:
                        if torch.cuda.is_available():
                            console.print(
                                "[yellow]CUDAExecutionProvider is not available in onnxruntime. Running on CPU.[/yellow]"
                            )
                        else:
                            console.print("[yellow]CUDA is not available. Running on CPU.[/yellow]")
                        return torch.device("cpu")
                case _:
                    raise ValueError(f"Unknown runtime config type {type(runtime_config)}")
        case _:
            raise ValueError(f"Invalid device {device}. Must be one of cpu or cuda")

    return device_chosen


# TODO: !! two seperate functions torch and onnx
def benchmark_model(
    model_cfg: BaseModelConfig,
    measure_memory: Optional[bool] = True,
    progress_bar: Optional[Progress] = None,
) -> pd.DataFrame:
    """Benchmark one model configuration.

    Parameters
    ----------
    model_cfg : :class:`~nvbenjo.cfg.TorchModelConfig` or :class:`~nvbenjo.cfg.OnnxModelConfig`
        Model configuration to benchmark
    measure_memory : bool, optional
        Whether to measure memory usage during benchmarking, by default True
    progress_bar : Progress, optional
        Progress bar instance to display benchmarking progress, by default None

    Returns
    -------
    pd.DataFrame
        Benchmarking results as a pandas DataFrame
    """
    results = []
    num_model_parameters = None
    precision_batch_oom = {}

    if progress_bar is None:
        progress_bar = _get_progress_bar()
    console = progress_bar.console

    iter_cfgs = list(itertools.product(*[model_cfg.devices, model_cfg.batch_sizes, model_cfg.runtime_options.items()]))
    bench_task = progress_bar.add_task("Running Benchmark", total=len(iter_cfgs))
    for device, batch_size, (runtime_option_name, runtime_cfg) in iter_cfgs:
        if precision_batch_oom.get(runtime_option_name, np.inf) < batch_size:
            # already went oom for these runtime options with smaller batch size -> skip bigger one
            progress_bar.advance(bench_task)
            continue
        try:
            device = _get_device(runtime_cfg, device, console)
            progress_bar.update(
                bench_task, description=f"  Device {device} | batch-size: {batch_size} | {runtime_option_name}"
            )

            model = load_model(model_cfg.type_or_path, device=device, runtime_config=runtime_cfg, **model_cfg.kwargs)
            if isinstance(model_cfg, TorchModelConfig):
                batch: utils.TensorLike
                batch, set_dtype = utils.get_rnd_from_shape_s(shape=model_cfg.shape, batch_size=batch_size)

                if num_model_parameters is None:
                    num_model_parameters = torch_utils.get_model_parameters(model)

                model = torch_utils.apply_non_amp_model_precision(
                    model, precision=model_cfg.runtime_options[runtime_option_name].precision
                )
                if runtime_cfg.compile:
                    model = torch.compile(model, **runtime_cfg.compile_kwargs)

                if not isinstance(model, nn.Module):
                    raise ValueError(f"Expected a torch.nn.Module but got {type(model)}")

                # only apply precision to input if no precision is specified
                if not set_dtype:
                    batch = torch_utils.apply_batch_precision(batch, precision=runtime_cfg.precision)
                else:
                    if not isinstance(batch, dict):
                        raise ValueError(f"Batch {batch} must be a dict if set_dtype is used.")
                    batch = ty.cast(
                        dict[str, torch.Tensor],
                        {
                            k: (
                                torch_utils.apply_batch_precision(
                                    ty.cast(torch.Tensor, v), precision=runtime_cfg.precision
                                )
                                if not set_dtype[k]
                                else ty.cast(torch.Tensor, v)
                            )
                            for k, v in batch.items()
                        },
                    )

                with torch_utils.get_amp_ctxt_for_precision(precision=runtime_cfg.precision, device=device):
                    _run_warmup(model, batch, device, model_cfg.num_warmup_batches, progress_bar)
                    if measure_memory:
                        memory_alloc = torch_utils.measure_memory_allocation(model, batch, device)
                    else:
                        memory_alloc = 0
                    if runtime_cfg.enable_profiling:
                        if "activities" not in runtime_cfg.profiler_kwargs:
                            runtime_cfg.profiler_kwargs["activities"] = (
                                [
                                    torch.profiler.ProfilerActivity.CPU,
                                    torch.profiler.ProfilerActivity.CUDA,
                                ]
                                if device.type == "cuda"
                                else [torch.profiler.ProfilerActivity.CPU]
                            )
                        profiler = torch.profiler.profile(
                            **runtime_cfg.profiler_kwargs,
                        )
                        profiler.start()
                    else:
                        profiler = None
                    cur_results = _measure_timings(
                        model=model,
                        batch=batch,
                        batch_size=batch_size,
                        device=device,
                        num_batches=model_cfg.num_batches,
                        progress_bar=progress_bar,
                        timing_function=torch_utils.measure_repeated_inference_timing,
                    )
                    if profiler is not None:
                        profiler.stop()
                        time_str = time.strftime("%Y-%m-%d_%H-%M-%S")
                        profiler.export_chrome_trace(
                            f"{runtime_cfg.profiling_prefix}_{device}_{batch_size}_{time_str}.json"
                        )
            elif isinstance(model_cfg, OnnxModelConfig):
                from nvbenjo import onnx_utils

                batch = onnx_utils.get_rnd_input_batch(model.get_inputs(), model_cfg.shape, batch_size)

                memory_alloc = 0
                num_model_parameters = 0
                set_dtype = False
                if measure_memory:
                    memory_alloc = onnx_utils.measure_memory_allocation(model, batch, device)
                else:
                    memory_alloc = 0
                cur_results = _measure_timings(
                    model,
                    batch,
                    batch_size,
                    device,
                    model_cfg.num_batches,
                    progress_bar,
                    timing_function=onnx_utils.measure_repeated_inference_timing,
                )
            else:
                raise ValueError(f"Unknown model config type {type(model_cfg)}")

            del model
            del batch
            torch.cuda.empty_cache()

            cur_results["memory_bytes"] = memory_alloc
            cur_results["model"] = model_cfg.name
            cur_results["batch_size"] = batch_size
            cur_results["runtime_options"] = runtime_option_name
            cur_results["device"] = str(device)
            results.append(cur_results)
        except torch.cuda.OutOfMemoryError:
            console.print(
                f"[red]Out of memory for batch size {batch_size} and runtime_options {runtime_option_name} on device {str(device)}[/red]"
            )
            precision_batch_oom[runtime_option_name] = batch_size
            continue
        except Exception as e:
            if "Failed to allocate memory" in str(e) or "ALLOC_FAILED" in str(e):
                console.print(
                    f"[red]Out of memory for batch size {batch_size} and runtime_options {runtime_option_name} on device {str(device)}[/red]"
                )
                precision_batch_oom[runtime_option_name] = batch_size
                continue
            else:
                raise e
        finally:
            progress_bar.advance(bench_task)

    return pd.concat(results)



import typing as ty
from dataclasses import dataclass

import pytest
import torch

try:
    from nvbenjo import onnx_utils
except ImportError as e:
    if "onnxruntime" in str(e):
        pytest.skip("onnxruntime is not installed, skipping ONNX utils tests.", allow_module_level=True)
    else:
        raise
from nvbenjo.cfg import OnnxRuntimeConfig


@dataclass
class FakeOnnxInput:
    name: str
    type: str
    shape: list[ty.Union[int, str]]


def test_get_model():
    with pytest.raises(
        ValueError, match="Invalid model tests/data/doesnotexist.onnx. Must be a valid ONNX path ending with .onnx"
    ):
        _ = onnx_utils.get_model(
            "tests/data/doesnotexist.onnx", device=torch.device("cpu"), runtime_config=OnnxRuntimeConfig()
        )


@pytest.mark.parametrize(
    "user_input_shapes",
    [
        ("B", 3, 224, 224),
        (("B", 3, 224, 224),),
        {"name": "input", "type": "float", "shape": ("B", 3, 224, 224)},
        {"name": "input", "type": "float"},
        {"name": "input"},
    ],
)
def test_get_rnd_input_batch(user_input_shapes):
    inputs = [FakeOnnxInput(name="input", type="tensor(float)", shape=["B", 3, 224, 224])]
    user_input_shapes = ("B", 3, 224, 224)
    batch_size = 4
    rnd_inputs = onnx_utils.get_rnd_input_batch(inputs, user_input_shapes, batch_size)
    assert isinstance(rnd_inputs, dict)
    assert len(rnd_inputs) == len(inputs)
    for inp_name, inp in rnd_inputs.items():
        assert isinstance(inp, torch.Tensor)
        assert inp.shape == (batch_size, 3, 224, 224)
        assert inp.dtype == torch.float32
        assert inp_name == "input"


def test_invalid_get_rnd_input_batch():
    inputs = [
        FakeOnnxInput(name="input1", type="tensor(float)", shape=["B", 3, 224, 224]),
        FakeOnnxInput(name="input2", type="tensor(int64)", shape=[1, 10]),
    ]
    batch_size = 4

    # single shape but multiple model inputs
    user_input_shapes = ("B", 3, 224, 224)
    with pytest.raises(ValueError, match="The model has multiple inputs, but the provided input is a single shape."):
        _ = onnx_utils.get_rnd_input_batch(inputs, user_input_shapes, batch_size)

    # mismatching number of shapes
    user_input_shapes = (("B", 3, 224, 224), ("B", 10), (1, 5))
    with pytest.raises(ValueError, match="The model has 2 inputs, but the provided input has 3 shapes."):
        _ = onnx_utils.get_rnd_input_batch(inputs, user_input_shapes, batch_size)

    # invalid input name
    user_input_shapes = (
        {"name": "invalid_input", "type": "float", "shape": ("B", 3, 224, 224)},
        {"name": "input2", "type": "int", "shape": (1, 10)},
    )
    with pytest.raises(ValueError, match="The model does not have an input named invalid_input."):
        _ = onnx_utils.get_rnd_input_batch(inputs, user_input_shapes, batch_size)

    # invalid num shapes
    user_input_shapes = ({"name": "input1", "type": "float", "shape": ("B", 3, 224, 2.3)},)
    with pytest.raises(ValueError, match="The model has 2 inputs, but the provided input has 1 shapes."):
        _ = onnx_utils.get_rnd_input_batch(inputs, user_input_shapes, batch_size)

    # invalid shape
    user_input_shapes = (
        {"name": "input1", "type": "float", "shape": ("B", 3, 224, 2.3)},
        {"name": "input2", "type": "float", "shape": ("B", 3, 224, 2.3)},
    )
    with pytest.raises(ValueError, match="Failed to generate random input from shape"):
        _ = onnx_utils.get_rnd_input_batch(inputs, user_input_shapes, batch_size)



import os
import time
import nvitop
import threading
import typing as ty

from .cfg import OnnxRuntimeConfig

try:
    import onnxruntime as ort  # type: ignore
except ImportError:
    raise ImportError("onnxruntime is not installed. Please install onnx and onnxruntime or onnxruntime-gpu.")

import pandas as pd
import torch
from omegaconf import DictConfig, ListConfig, OmegaConf

from . import console
from .torch_utils import transfer_to_device
from .utils import EXAMPLE_VALID_SHAPES, TRANSFER_WARNING, _check_shape_dict, get_rnd_from_shape_s, Shape


def get_model(
    type_or_path: str, device: torch.device, runtime_config: OnnxRuntimeConfig, **kwargs
) -> ort.InferenceSession:
    """Load a onnx model into InferenceSession.

    Parameters
    ----------
    type_or_path : str
        Path to the ONNX model file.
    device : torch.device
        Device to run the ONNX model on.
    runtime_config : OnnxRuntimeConfig
        Runtime configuration for the ONNX model.
    **kwargs
        Additional keyword arguments to pass to the InferenceSession.

    Returns
    -------
    ort.InferenceSession
        Loaded ONNX InferenceSession.
    """
    type_or_path = os.path.expanduser(type_or_path)
    if not type_or_path.endswith(".onnx") or not os.path.isfile(type_or_path):
        raise ValueError(f"Invalid model {type_or_path}. Must be a valid ONNX path ending with .onnx")

    if runtime_config.execution_providers is None:
        if device.type == "cuda":
            if "CUDAExecutionProvider" not in ort.get_available_providers():  # type: ignore
                raise RuntimeError(
                    "CUDAExecutionProvider is not available in onnxruntime. Please install onnxruntime-gpu or run on CPU."
                )
            else:
                providers = [("CUDAExecutionProvider", {"device_id": device.index}), "CPUExecutionProvider"]
        else:
            providers = ["CPUExecutionProvider"]
    else:
        providers = runtime_config.execution_providers

    # Convert list in providers to tuple if needed
    if isinstance(providers, ListConfig):
        providers = OmegaConf.to_container(providers)
    for i, provider in enumerate(providers):  # type: ignore
        if isinstance(provider, (list, ListConfig)):
            providers[i] = tuple(provider)  # type: ignore

    session_options = ort.SessionOptions()  # type: ignore
    session_options.log_severity_level = runtime_config.log_severity_level
    session_options.intra_op_num_threads = runtime_config.intra_op_num_threads
    session_options.inter_op_num_threads = runtime_config.inter_op_num_threads
    session_options.enable_profiling = runtime_config.enable_profiling
    if runtime_config.enable_profiling and runtime_config.profiling_prefix is not None:
        session_options.profile_file_prefix = runtime_config.profiling_prefix
    session_options.graph_optimization_level = getattr(
        ort.GraphOptimizationLevel,  # type: ignore
        runtime_config.graph_optimization_level,
    )

    sess = ort.InferenceSession(
        type_or_path,
        sess_options=session_options,
        providers=providers,  # type: ignore
        provider_options=runtime_config.provider_options,
        **kwargs,
    )
    return sess


def _sample_gpu_memory(
    device: torch.device, stop_event: threading.Event, max_mem: list[int], sample_time_s: float = 0.010
):
    if device.type == "cuda":
        gpu = nvitop.Device(device.index)
    else:
        max_mem[0] = 1
    while not stop_event.is_set():
        if device.type == "cuda":
            mem = gpu.memory_used()
            if isinstance(mem, int) and mem > max_mem[0]:
                max_mem[0] = mem
        time.sleep(sample_time_s)


def measure_memory_allocation(
    model: ort.InferenceSession, sample: dict[str, torch.Tensor], device: torch.device, iterations: int = 3
) -> int:
    """Measure the memory usage during inference

    Parameters
    ----------
    model : ort.InferenceSession
        The ONNX model to benchmark.
    sample : dict[str, torch.Tensor]
        Sample input data for the model.
    device : torch.device
        Device to run the model on.
    iterations : int, optional
        Number of iterations to run for measuring memory allocation, by default 3

    Returns
    -------
    int
        Maximum memory allocated during inference in bytes.
    """
    max_mem = [-1]
    stop_event = threading.Event()

    # Start memory sampling thread so we can measure peak memory usage in parallel
    sampler = threading.Thread(target=_sample_gpu_memory, args=(device, stop_event, max_mem))
    sampler.start()
    time.sleep(0.01)  # give sampler some time to start

    try:
        for _ in range(iterations):
            _ = model.run(None, {n: d.cpu().numpy() for n, d in sample.items()})
    finally:
        stop_event.set()
        sampler.join()

    max_mem = max_mem[0]
    if max_mem == -1:
        raise RuntimeError("Memory measurement failed!")

    return max_mem


def _get_formated_input_info(onnx_inputs) -> str:
    return "\n".join(
        f"- name: {onnx_input.name}, dtype: {onnx_input.type}, shape: {onnx_input.shape}{', default: ' + str(onnx_input.value) if hasattr(onnx_inputs, 'value') else ''}"
        for onnx_input in onnx_inputs
    )


def get_rnd_input_batch(onnx_session_inputs, shape: Shape, batch_size: int) -> dict[str, torch.Tensor]:
    def strip_type_string(s: str) -> str:
        if s.startswith("tensor(") and s.endswith(")"):
            s = s[len("tensor(") : -1]
        return s

    if not isinstance(shape, dict) and all(isinstance(si, (str, int)) for si in shape):
        # simple shape e.g. (B, 3, 224, 224)
        if len(onnx_session_inputs) != 1:
            raise ValueError(
                "The model has multiple inputs, but the provided input is a single shape."
                f"Model Inputs: \n{_get_formated_input_info(onnx_session_inputs)}"
            )
        model_input = onnx_session_inputs[0]
        rnd_shape = ({"name": model_input.name, "type": strip_type_string(model_input.type), "shape": shape},)
    elif all(isinstance(si, (tuple, list, ListConfig)) for si in shape):
        # tuple of shapes e.g. ((B, 3, 224, 224), (B, 10))
        if len(onnx_session_inputs) != len(shape):
            raise ValueError(
                f"The model has {len(onnx_session_inputs)} inputs, but the provided input has {len(shape)} shapes. Please provide a list of shapes or a dict of shapes."
                f"Model Inputs: \n{_get_formated_input_info(onnx_session_inputs)}"
            )
        rnd_shape = tuple(
            {"name": model_input.name, "type": strip_type_string(model_input.type), "shape": si}
            for model_input, si in zip(onnx_session_inputs, shape)
        )
    elif all(isinstance(si, (dict, DictConfig)) for si in shape):
        onnx_inputs_by_name = {inp.name: inp for inp in onnx_session_inputs}
        rnd_shape = tuple(dict(si) for si in shape)  # convert from DictConfig to dict # type: ignore
        if len(onnx_session_inputs) != len(shape):
            raise ValueError(
                f"The model has {len(onnx_session_inputs)} inputs, but the provided input has {len(shape)} shapes. Please provide a list of shapes or a dict of shapes."
                f"Model Inputs: \n{_get_formated_input_info(onnx_session_inputs)}"
            )
        for si in rnd_shape:
            _check_shape_dict(si)
            # # name='input', type='float', shape=['B', 320000, 1], min_max=(0, 1)
            if si["name"] not in onnx_inputs_by_name:
                raise ValueError(
                    f"The model does not have an input named {si['name']}."
                    f"Model Inputs: \n{_get_formated_input_info(onnx_session_inputs)}"
                )
            if "type" not in si:
                si["type"] = strip_type_string(onnx_inputs_by_name[si["name"]].type)
            if "shape" not in si:
                si["shape"] = onnx_inputs_by_name[si["name"]].shape
    else:
        raise ValueError(
            (
                f"Invalid shape {shape}.\n "
                "Example valid inputs:\n " + "\n - ".join([str(s) for s in EXAMPLE_VALID_SHAPES])
            )
        )
    batch, _ = get_rnd_from_shape_s(shape=rnd_shape, batch_size=batch_size)
    if not isinstance(batch, dict):
        raise ValueError("Internal Error was unable to generate dict of inputs for ONNX model.")
    return batch  # type: ignore


def measure_repeated_inference_timing(
    model: ort.InferenceSession,
    sample: dict[str, torch.Tensor],
    batch_size: int,
    model_device: torch.device,
    transfer_to_device_fn: ty.Callable = transfer_to_device,
    num_runs: int = 100,
    progress_callback: ty.Optional[ty.Callable] = None,
) -> pd.DataFrame:
    """Measure inference times.

    Parameters
    ----------
    model : ort.InferenceSession
        The ONNX model to benchmark.
    sample : dict[str, torch.Tensor]
        Sample input data for the model.
    batch_size : int
        Batch size for the input data.
    model_device : torch.device
        Device to run the model on.
    transfer_to_device_fn : Callable, optional
        Function to transfer data to the device, by default transfer_to_device
    num_runs : int, optional
        Number of runs to perform for timing, by default 100
    progress_callback : ty.Optional[ty.Callable], optional
        Callback function for p

    Returns
    -------
    pd.DataFrame
        DataFrame containing timing results.
    """
    time_cpu_to_device = []
    time_inference = []
    time_device_to_cpu = []
    time_total = []
    results_raw = []

    # Convert ONNX type to PyTorch dtype
    dtype_map = {
        "tensor(float)": torch.float32,
        "tensor(float16)": torch.float16,
        "tensor(int64)": torch.int64,
        "tensor(int32)": torch.int32,
        "tensor(bool)": torch.bool,
    }

    onnx_model_outputs = model.get_outputs()

    # run model once so we know the output shapes
    outputs = model.run(None, {n: d.cpu().numpy() for n, d in sample.items()})
    output_shapes = {
        onnx_output.name: output_shape.shape for onnx_output, output_shape in zip(onnx_model_outputs, outputs)
    }
    del outputs

    for _ in range(num_runs):
        start_on_cpu = time.perf_counter()
        device_sample = transfer_to_device_fn(sample, model_device)

        if model_device.type == "cuda":
            start_event = torch.cuda.Event(enable_timing=True)
            stop_event = torch.cuda.Event(enable_timing=True)
            start_event.record()  # For GPU timing
        start_on_device = time.perf_counter()  # For CPU timing

        io_binding = model.io_binding()
        device_id = 0 if model_device.index is None else model_device.index

        if isinstance(device_sample, dict):
            for i, (name, input) in enumerate(device_sample.items()):
                io_binding.bind_input(
                    name=name,
                    device_type=model_device.type,
                    device_id=device_id,
                    element_type=str(input.dtype).strip("torch."),
                    shape=input.shape,
                    buffer_ptr=input.data_ptr(),
                )
        else:
            raise ValueError(f"Invalid input type {type(device_sample)}. Must be one of list, tuple, dict")

        device_result = []
        for i, output in enumerate(onnx_model_outputs):
            torch_dtype = dtype_map.get(output.type, torch.float32)  # default to float32 if type not found
            output_tensor = torch.empty(size=output_shapes[output.name], dtype=torch_dtype, device=model_device)
            io_binding.bind_output(
                name=output.name,
                device_type=model_device.type,
                device_id=device_id,
                element_type=str(output_tensor.dtype).strip("torch."),
                shape=output_tensor.shape,
                buffer_ptr=output_tensor.data_ptr(),
            )
            device_result.append(output_tensor)

        model.run_with_iobinding(io_binding)

        if model_device.type == "cuda":
            stop_event.record()
            torch.cuda.synchronize()
            elapsed_on_device = start_event.elapsed_time(stop_event) / 1000.0
            stop_on_device = time.perf_counter()
        else:
            stop_on_device = time.perf_counter()
            elapsed_on_device = stop_on_device - start_on_device

        try:
            transfer_to_device_fn(device_result, torch.device("cpu"))
        except Exception:
            console.print(TRANSFER_WARNING)
        stop_on_cpu = time.perf_counter()

        assert elapsed_on_device > 0

        time_cpu_to_device.append(start_on_device - start_on_cpu)
        time_inference.append(elapsed_on_device)
        time_device_to_cpu.append(stop_on_cpu - stop_on_device)
        time_total.append(stop_on_cpu - start_on_cpu)
        results_raw.append(
            {
                "time_cpu_to_device": start_on_device - start_on_cpu,
                "time_inference": elapsed_on_device,
                "time_device_to_cpu": stop_on_cpu - stop_on_device,
                "time_total": stop_on_cpu - start_on_cpu,
                "time_total_batch_normalized": (stop_on_cpu - start_on_cpu) / batch_size,
            }
        )
        if progress_callback is not None:
            progress_callback()

    results_raw = pd.DataFrame(results_raw)

    return results_raw



import logging
import os
import sys
import typing as ty
from importlib.resources import files
from os.path import join

import hydra
from hydra.core.config_store import ConfigStore
from omegaconf import DictConfig, OmegaConf
from rich.logging import RichHandler

from . import console, plot
from .benchmark import benchmark_models
from .cfg import BenchConfig, instantiate_model_configs
from .system_info import get_system_info

logger = logging.getLogger(__name__)

cs = ConfigStore.instance()
cs.store(name="base_config", node=BenchConfig)


@hydra.main(version_base=None, config_path=os.path.join(files("nvbenjo").joinpath("conf")), config_name="default")
def _run_nvbenjo(cfg: ty.Union[BenchConfig, DictConfig]):
    run(cfg)


def run(cfg: ty.Union[BenchConfig, DictConfig]) -> None:
    logging.basicConfig(level="NOTSET", format="%(message)s", datefmt="[%X]", handlers=[RichHandler(console=console)])
    models = instantiate_model_configs(cfg)
    if cfg.output_dir is not None:
        output_dir = os.path.abspath(cfg.output_dir)

    system_info = get_system_info()

    if cfg.output_dir is not None:
        logger.info(f"Starting benchmark, output-dir {output_dir}")

    if len(models) == 0:
        logger.info("No models to benchmark, please specify a configuration or override via the command line.")
        return
    results = benchmark_models(models, measure_memory=cfg.nvbenjo.measure_memory)

    if cfg.output_dir is not None:
        results.to_csv(join(output_dir, "out.csv"))
        with open(join(output_dir, "config.yaml"), "w") as f:
            f.write(OmegaConf.to_yaml(cfg))

    custom_metric_keys = list(set(sum([list(mcfg.custom_batchmetrics.keys()) for mcfg in models.values()], [])))
    if cfg.output_dir is not None:
        logger.info("Generating plots...")
        plot.visualize_results(
            results,
            keys=[
                "time_cpu_to_device",
                "time_device_to_cpu",
                "time_inference",
                "time_total_batch_normalized",
                "memory_bytes",
            ]
            + custom_metric_keys,
            output_dir=output_dir,
        )
    plot.print_system_info(system_info)
    plot.print_results(results, custom_metric_keys=custom_metric_keys)
    logger.info(f"Benchmark finished, outputs in: {output_dir}")


def _fix_config_path():
    # NOTE: this is a workaround to allow specifying config file with full path
    #       since hydra only allows config name and config dir
    #       so for -cn /path/to/config.yaml we add -cd /path/to and change -cn to config.yaml
    if "-cn" in sys.argv or "--config-name" in sys.argv and "-cd" not in sys.argv and "--config-dir" not in sys.argv:
        arg_index = sys.argv.index("-cn") if "-cn" in sys.argv else sys.argv.index("--config-name")  # type: ignore
        cfg_index = arg_index + 1
        if cfg_index <= len(sys.argv) - 1:
            config_name = sys.argv[cfg_index]
            if os.path.dirname(config_name):
                sys.argv.append("-cd")
                sys.argv.append(os.path.dirname(config_name))
                sys.argv[cfg_index] = os.path.basename(config_name)
                logger.debug("Sys argv: " + str(sys.argv))
                logger.debug(
                    f"Adjusted config path, using -cd {os.path.dirname(config_name)} and -cn {os.path.basename(config_name)}"
                )


def nvbenjo():
    _fix_config_path()
    _run_nvbenjo()


if __name__ == "__main__":
    nvbenjo()



from __future__ import annotations

import typing as ty
import pandas as pd
from enum import Enum
from omegaconf.listconfig import ListConfig
from omegaconf.dictconfig import DictConfig

import torch

BATCH_SIZE_IDENTIFIERS = ("B", "batch_size")
TRANSFER_WARNING = (
    "[yellow]Warning: Could not transfer model output to CPU. Time to CPU measures will be incorrect.[/yellow]"
)

SingleShape = tuple[int | str, ...] | dict[str, ty.Any]
MultiShape = tuple[SingleShape, ...]
Shape = SingleShape | MultiShape
TensorLike = torch.Tensor | tuple[torch.Tensor, ...] | dict[str, torch.Tensor]
ProviderType = ty.Union[str, ty.Tuple[str, dict[ty.Any, ty.Any]]]

EXAMPLE_VALID_SHAPES: list[Shape] = [
    ("B", 3, 224, 224),
    (("B", 3, 224, 224), ("B", 10)),
    ({"name": "input1", "type": "float", "shape": ("B", 3, 224, 224), "min_max": (0, 1)},),
    (
        {"name": "input1", "type": "float", "shape": ("B", 3, 224, 224), "min_max": (0, 1)},
        {"name": "input2", "type": "int", "shape": (1, 3)},
        {"name": "input3", "type": "int", "shape": (), "value": 42},
    ),
]

AMP_PREFIX = "amp"


class NoBatchShapeError(ValueError):
    pass


class PrecisionType(Enum):
    AMP = f"{AMP_PREFIX}"
    AMP_FP16 = f"{AMP_PREFIX}_fp16"
    AMP_BFLOAT16 = f"{AMP_PREFIX}_bfloat16"
    FP32 = "fp32"
    FP16 = "fp16"
    BFLOAT16 = "bfloat16"
    LONG = "long"


def format_num(num: ty.Union[int, float], bytes: bool = False) -> ty.Union[str, None]:
    """Scale bytes to its proper format, e.g. 1253656 => '1.20MB'"""
    if num is None:
        return num
    factor = 1024 if bytes else 1000
    suffix = "B" if bytes else ""
    for unit in ["", " K", " M", " G", " T", " P"]:
        if num < factor:
            return f"{num:.2f}{unit}{suffix}"
        num /= factor


def format_seconds(time_seconds: float) -> str:
    if time_seconds > 1:
        return f"{time_seconds:.3f} s"
    else:
        time_ms = time_seconds * 1000
        if time_ms > 1:
            return f"{time_ms:.3f} ms"
        else:
            time_us = time_ms * 1000
            return f"{time_us:.3f} us"


def _get_rnd(
    shape_tuple: tuple[int], dtype: ty.Optional[str], min_val: int, max_val: int, value: ty.Optional[ty.Any] = None
) -> torch.Tensor:
    if dtype is None or any(s in dtype for s in ["float", "double"]):
        dtype = dtype if dtype is not None else "float32"
        if value:
            if isinstance(value, (ListConfig, list)):
                rnd = torch.tensor(value, dtype=getattr(torch, dtype))
            else:
                rnd = torch.full(size=shape_tuple, fill_value=value, dtype=getattr(torch, dtype))
        else:
            rnd = torch.distributions.Uniform(min_val, max_val).sample(shape_tuple).to(dtype=getattr(torch, dtype))
    elif any(s in dtype for s in ["int", "long"]):
        if value:
            if isinstance(value, (ListConfig, list)):
                rnd = torch.tensor(value, dtype=getattr(torch, dtype))
            else:
                rnd = torch.full(size=shape_tuple, fill_value=value, dtype=getattr(torch, dtype))
        else:
            rnd = torch.randint(low=int(min_val), high=int(max_val), size=shape_tuple, dtype=getattr(torch, dtype))
    else:
        raise ValueError(f"Invalid dtype {dtype}. Must be one of int, long, float, double")
    return rnd


def _check_shape_dict(si: dict | DictConfig) -> None:
    if "name" not in si:
        raise ValueError(f"The shape definition {si} must contain a 'name' key.")
    if "dtype" in si:
        raise ValueError("The 'dtype' is not valid. Did you mean 'type'?")
    if "shape" not in si:
        raise ValueError(f"The shape definition {si} must contain a 'shape' key.")
    for k in si.keys():
        if k not in ["name", "type", "shape", "min_max", "value"]:
            raise ValueError(f"Invalid key {k} in shape definition {si}.")
    if "min_max" in si.keys() and "value" in si.keys():
        raise ValueError(f"Invalid shape definition {si} can only specify min_max or value.")


def get_rnd_from_shape_s(
    shape: Shape, batch_size: int, dtype=None, min_val=0, max_val=1
) -> tuple[TensorLike, dict[str, bool]]:
    try:
        depends_on_batch = False
        set_individual_types = {}

        def _get_rnd_batch(shape_tuple):
            if any(s in shape_tuple for s in BATCH_SIZE_IDENTIFIERS):
                nonlocal depends_on_batch
                depends_on_batch = True
            return tuple(batch_size if s in BATCH_SIZE_IDENTIFIERS else s for s in shape_tuple)

        if not isinstance(shape, dict) and all(isinstance(si, (str, int)) for si in shape):
            # simple shape e.g. (B, 3, 224, 224)
            rnd_input = _get_rnd(shape_tuple=_get_rnd_batch(shape), dtype=dtype, min_val=min_val, max_val=max_val)
        elif all(isinstance(si, (tuple, list, ListConfig)) for si in shape):
            # tuple of shapes e.g. ((B, 3, 224, 224), (B, 10))
            rnd_input = tuple(
                _get_rnd(_get_rnd_batch(si), dtype=dtype, min_val=min_val, max_val=max_val) for si in shape
            )
        elif all(isinstance(si, (dict, DictConfig)) for si in shape):
            rnd_input = {}
            for si in shape:
                # check is redundant but helps type checker
                if not isinstance(si, (dict, DictConfig)):
                    raise ValueError(f"Shape item {si} must be of type dict.")

                # name='input', type='float', shape=['B', 320000, 1], min_max=(0, 1)
                _check_shape_dict(si)
                name = si["name"]
                set_individual_types[name] = "type" in si

                rnd_input[name] = _get_rnd(
                    shape_tuple=_get_rnd_batch(si["shape"]),
                    dtype=si.get("type", dtype),
                    min_val=si.get("min_max", (min_val, max_val))[0],
                    max_val=si.get("min_max", (min_val, max_val))[1],
                    value=si.get("value", None),
                )
        else:
            raise ValueError(
                (
                    f"Invalid shape {shape}.\n "
                    "Example valid inputs:\n " + "\n - ".join([str(s) for s in EXAMPLE_VALID_SHAPES])
                )
            )

        if not depends_on_batch:
            raise NoBatchShapeError(
                f"Shape {shape} does not depend on batch size. "
                f"Please ensure that the shape contains an identifier for the batch size: {BATCH_SIZE_IDENTIFIERS}."
            )
    except NoBatchShapeError as e:
        raise e
    except Exception as e:
        raise ValueError(
            f"Failed to generate random input from shape {shape} with batch size {batch_size}. "
            f"Please ensure that the shape is valid.\n"
            + "Example valid inputs:\n "
            + "\n - ".join([str(s) for s in EXAMPLE_VALID_SHAPES])
        ) from e

    return rnd_input, set_individual_types


def calculate_batchmetrics(results: pd.DataFrame, custom_batchmetrics: dict[str, float]) -> pd.DataFrame:
    """Calculate custom batch metrics and add them to the results DataFrame.

    Parameters
    ----------
    results : pd.DataFrame
        The benchmark results DataFrame.
    custom_batchmetrics : dict[str, float]
        Dictionary of custom batch metrics to calculate. The key is the metric name and the value is the multiplier.

    Returns
    -------
    pd.DataFrame
        The updated results DataFrame with custom batch metrics added.
    """
    for metric_name, value in custom_batchmetrics.items():
        results[metric_name] = value / results["time_total_batch_normalized"]
    return results



import os
import typing as ty
from abc import ABC
from contextlib import nullcontext
from dataclasses import dataclass, field

from hydra.utils import instantiate
from omegaconf import DictConfig, OmegaConf, open_dict

from .utils import PrecisionType, ProviderType


@dataclass
class BaseModelConfig(ABC):
    """Base model configuration

    Parameters
    ----------
    name : str
        Name of the model.
    type_or_path : str
        Model type or path. Can be a local file path or a model identifier.
    kwargs : dict
        Additional keyword arguments to pass when instantiating the model.
    shape : tuple
        Input shape of the model. Use "B" or "batch_size" to denote the batch size dimension.
    num_warmup_batches : int
        Number of warm-up batches to run before measuring performance.
    num_batches : int
        Number of batches to run for performance measurement.
    batch_sizes : tuple
        Tuple of batch sizes to benchmark.
    devices : tuple of str
        Tuple of device names to benchmark on (e.g., 'cpu', 'cuda:0').
    runtime_options : dict[str, ty.Any]
        Dictionary mapping runtime names to their specific runtime configurations.
    custom_batchmetrics : dict[str, float]
        Dictionary of custom batch metrics to include in the benchmark results.
        These are calculated as `value / time_total_batch_normalized`.
        For example if each batch is a frame you can calculate FPS as:
            custom_batchmetrics:
                fps: 1.0
                batch-per-second: 1.0
        or if each batch is 3s of audio you can calculate the real time factor as:
            custom_batchmetrics:
                real-time-factor: 1/3
    """

    name: str = "resnet"
    type_or_path: str = "torchvision:wide_resnet101_2"
    kwargs: dict = field(default_factory=dict)
    shape: tuple = ("B", 3, 224, 224)
    num_warmup_batches: int = 5
    num_batches: int = 50
    batch_sizes: tuple = (16, 32)
    devices: tuple[str] = ("cpu",)
    runtime_options: dict[str, ty.Any] = field(default_factory=dict)
    custom_batchmetrics: dict[str, float] = field(default_factory=dict)


@dataclass
class NvbenjoConfig:
    """
    Root configuration for nvbenjo benchmarking.

    Parameters
    ----------
    measure_memory : bool
        Whether to measure GPU memory allocation during benchmarking.
    models: dict[str, TorchModelConfig | OnnxModelConfig]
        Dictionary mapping model names to their configurations.
        See :class:`TorchModelConfig` and :class:`OnnxModelConfig` for details.
    """

    measure_memory: bool = True
    models: dict[str, ty.Any] = field(default_factory=lambda: dict())


@dataclass
class BenchConfig:
    """
    Main benchmark configuration container.

    Parameters
    ----------
    nvbenjo : NvbenjoConfig
        Nvbenjo-specific configuration settings.
    output_dir : str or None
        Directory path where benchmark results will be saved.
        If None, uses Hydra's default output directory.
    """

    nvbenjo: NvbenjoConfig = field(default_factory=NvbenjoConfig)
    output_dir: ty.Optional[str] = None


@dataclass
class TorchRuntimeConfig:
    """PyTorch Runtime configuration:

    Parameters
    ----------
    compile : bool
        Whether to compile the model using torch.compile (PyTorch 2.0+).
    compile_kwargs : dict
        Additional keyword arguments for torch.compile.
    precision : PrecisionType
        Precision type for model inference (e.g., fp32, fp16, amp).
    enable_profiling : bool
        Whether to enable PyTorch profiler during inference.
    profiling_prefix : str or None
        Prefix for profiler output files. If None, a default path will be used.
    profiler_kwargs : dict
        Additional keyword arguments for torch.profiler.profile.
    """

    compile: bool = False
    compile_kwargs: dict = field(default_factory=dict)
    precision: PrecisionType = PrecisionType.FP32
    enable_profiling: bool = False
    profiling_prefix: ty.Optional[str] = None
    profiler_kwargs: dict = field(default_factory=dict)


@dataclass
class OnnxRuntimeConfig:
    """ONNX Runtime configuration:

    Parameters
    ----------
    execution_providers : tuple of str or None
        Tuple of execution providers to use (e.g., ('CPUExecutionProvider',
        'CUDAExecutionProvider')). If None, uses the default provider.
    graph_optimization_level : str
        Graph optimization level for ONNX Runtime. Options are 'ORT_ENABLE_ALL', 'ORT_ENABLE_LAYOUT',
        'ORT_ENABLE_BASIC', 'ORT_DISABLE_ALL'.
    intra_op_num_threads : int
        Number of threads used to parallelize the execution within nodes.
    inter_op_num_threads : int
        Number of threads used to parallelize the execution of the graph (between nodes)
    log_severity_level : int
        Logging severity level (0=VERBOSE, 1=INFO, 2=WARNING, 3=ERROR, 4=FATAL)
    enable_profiling : bool
        Whether to enable profiling in ONNX Runtime.
    profiling_prefix : str or None
        Prefix for profiling output files. If None, a default path will be used.
    provider_options : sequence of dict or None
        Additional options for each execution provider.
    """

    execution_providers: ty.Optional[ty.List[ProviderType]] = None
    graph_optimization_level: str = (
        "ORT_ENABLE_ALL"  # 99 ORT_ENABLE_ALL, 3 ORT_ENABLE_LAYOUT, 1 ORT_ENABLE_BASIC, 0 ORT_DISABLE_ALL
    )
    intra_op_num_threads: int = 1
    inter_op_num_threads: int = 0
    log_severity_level: int = 2  # Warning
    enable_profiling: bool = False
    profiling_prefix: ty.Optional[str] = None
    provider_options: ty.Sequence[dict[ty.Any, ty.Any]] | None = None


@dataclass
class TorchModelConfig(BaseModelConfig):
    """PyTorch model configuration

    Parameters
    ----------
    name : str
        Name of the model.
    type_or_path : str
        Model type or path. Can be a local file path or a model identifier.
    kwargs : dict
        Additional keyword arguments to pass when instantiating the model.
    shape : tuple
        Input shape of the model. Use "B" to denote the batch size dimension.
        Examples::

            # Single input shape
            ("B", 3, 224, 224)

            # Multiple input shapes
            (("B", 3, 224, 224), ("B", 10))

            # Dictionary with metadata
            ({"name": "input1", "type": "float", "shape": ("B", 3, 224, 224), "min_max": (0, 1)},)

            # Multiple dictionary inputs
            (
                {"name": "input1", "type": "float", "shape": ("B", 3, 224, 224), "min_max": (0, 1)},
                {"name": "input2", "type": "int", "shape": (1, 3)},
                {"name": "input3", "type": "int", "shape": (), "value": 42},
            )
    num_warmup_batches : int
        Number of warm-up batches to run before measuring performance.
    num_batches : int
        Number of batches to run for performance measurement.
    batch_sizes : tuple
        Tuple of batch sizes to benchmark.
    devices : tuple of str
        Tuple of device names to benchmark on (e.g., 'cpu', 'cuda:0').
    runtime_options : dict[str, :class:`~nvbenjo.cfg.TorchRuntimeConfig`]
        Dictionary mapping runtime names to their specific runtime configurations.
    """

    model_kwargs: dict = field(default_factory=dict)
    runtime_options: dict[str, TorchRuntimeConfig] = field(default_factory=lambda: {"default": TorchRuntimeConfig()})

    def __post_init__(self):
        for i, (key, opt) in enumerate(self.runtime_options.items()):
            if isinstance(opt, DictConfig):
                self.runtime_options[key] = OmegaConf.structured(TorchRuntimeConfig(**OmegaConf.to_container(opt)))  # type: ignore


@dataclass
class OnnxModelConfig(BaseModelConfig):
    """ONNX model configuration

    Parameters
    ----------
    name : str
        Name of the model.
    type_or_path : str
        Model type or path. Can be a local file path or a model identifier.
    kwargs : dict
        Additional keyword arguments to pass when instantiating the model.
    shape : tuple
        Input shape of the model. Use "B" to denote the batch size dimension.

        Examples::

            # Single input shape
            ("B", 3, 224, 224)

            # Multiple input shapes
            (("B", 3, 224, 224), ("B", 10))

            # Dictionary with metadata
            ({"name": "input1", "type": "float", "shape": ("B", 3, 224, 224), "min_max": (0, 1)},)

            # Multiple dictionary inputs
            (
                {"name": "input1", "type": "float", "shape": ("B", 3, 224, 224), "min_max": (0, 1)},
                {"name": "input2", "type": "int", "shape": (1, 3)},
                {"name": "input3", "type": "int", "shape": (), "value": 42},
            )
    num_warmup_batches : int
        Number of warm-up batches to run before measuring performance.
    num_batches : int
        Number of batches to run for performance measurement.
    batch_sizes : tuple
        Tuple of batch sizes to benchmark.
    devices : tuple of str
        Tuple of device names to benchmark on (e.g., 'cpu', 'cuda:0').
    runtime_options : dict[str, :class:`~nvbenjo.cfg.OnnxRuntimeConfig`]
        Dictionary mapping runtime names to their specific runtime configurations.
    """

    runtime_options: dict[str, OnnxRuntimeConfig] = field(default_factory=lambda: {"default": OnnxRuntimeConfig()})

    def __post_init__(self):
        for i, (key, opt) in enumerate(self.runtime_options.items()):
            if isinstance(opt, DictConfig):
                self.runtime_options[key] = OmegaConf.structured(OnnxRuntimeConfig(**OmegaConf.to_container(opt)))  # type: ignore


def instantiate_model_configs(cfg: ty.Union[BenchConfig, DictConfig]) -> dict[str, BaseModelConfig]:
    models = {}
    runtimes = {}
    for model_name, model in cfg.nvbenjo.models.items():
        ctxt = open_dict(model) if isinstance(model, DictConfig) else nullcontext()
        if "_target_" not in model:
            with ctxt:
                if model["type_or_path"].endswith(".onnx"):
                    cfg.nvbenjo.models[model_name]["_target_"] = (
                        f"{OnnxModelConfig.__module__}.{OnnxModelConfig.__qualname__}"
                    )
                    cfg.nvbenjo.models[model_name]["_convert_"] = "all"
                    if "runtime_options" in model:
                        runtimes[model_name] = {}
                        for runtime_name in model["runtime_options"].keys():
                            cfg.nvbenjo.models[model_name]["runtime_options"][runtime_name]["_target_"] = (
                                f"{OnnxRuntimeConfig.__module__}.{OnnxRuntimeConfig.__qualname__}"
                            )
                            runtimes[model_name][runtime_name] = instantiate(
                                cfg.nvbenjo.models[model_name]["runtime_options"][runtime_name]
                            )
                else:
                    cfg.nvbenjo.models[model_name]["_target_"] = (
                        f"{TorchModelConfig.__module__}.{TorchModelConfig.__qualname__}"
                    )
                    cfg.nvbenjo.models[model_name]["_convert_"] = "all"
                    if "runtime_options" in model:
                        runtimes[model_name] = {}
                        for runtime_name in model["runtime_options"].keys():
                            cfg.nvbenjo.models[model_name]["runtime_options"][runtime_name]["_target_"] = (
                                f"{TorchRuntimeConfig.__module__}.{TorchRuntimeConfig.__qualname__}"
                            )
                            runtimes[model_name][runtime_name] = instantiate(
                                cfg.nvbenjo.models[model_name]["runtime_options"][runtime_name]
                            )
                            runtimes[model_name][runtime_name].precision = PrecisionType[
                                cfg.nvbenjo.models[model_name]["runtime_options"][runtime_name]["precision"].upper()
                            ]

        models[model_name] = instantiate(model) if isinstance(model, DictConfig) else model
        if model_name in runtimes:
            models[model_name].runtime_options = runtimes[model_name]

    # For onnx profiling we add a valid profiling prefix in the output directory if needed
    for model_name, model in models.items():
        if isinstance(model, (OnnxModelConfig, TorchModelConfig)):
            for runtime_name, runtime in model.runtime_options.items():
                if runtime.enable_profiling:
                    if runtime.profiling_prefix is None:
                        runtime.profiling_prefix = os.path.join(
                            cfg.output_dir, model_name, f"{model_name}_{runtime_name}_profile"
                        )
                    else:
                        # make sure the relative path is inside the output dir
                        if not os.path.abspath(runtime.profiling_prefix) == runtime.profiling_prefix:
                            runtime.profiling_prefix = os.path.abspath(
                                os.path.join(cfg.output_dir, runtime.profiling_prefix)
                            )
                    os.makedirs(os.path.dirname(runtime.profiling_prefix), exist_ok=True)

    return models



from nvbenjo.system_info import get_system_info


def test_get_system_info():
    system_info = get_system_info()
    assert isinstance(system_info, dict)
    assert "cpu" in system_info
    assert "gpus" in system_info
    assert "os" in system_info

    assert isinstance(system_info["gpus"], (dict, list))
    if isinstance(system_info["gpus"], list):
        for gpu in system_info["gpus"]:
            assert isinstance(gpu, dict)
            assert "name" in gpu
            assert "architecture" in gpu
            assert "memory" in gpu
            assert "cuda_capability" in gpu
            assert "driver" in gpu

    # check basic cpu info
    cpu_info = system_info["cpu"]
    assert isinstance(cpu_info, dict)
    assert "cores" in cpu_info
    assert "frequency" in cpu_info
    assert "model" in cpu_info
    assert "architecture" in cpu_info



from hydra.core.config_search_path import ConfigSearchPath
from hydra.plugins.search_path_plugin import SearchPathPlugin


class NvbenjoSearchPathPlugin(SearchPathPlugin):
    def manipulate_search_path(self, search_path: ConfigSearchPath) -> None:
        search_path.append(provider="nvbenjo-plugin", path="pkg://nvbenjo/conf")
        # allow from current directory
        search_path.append(provider="nvbenjo-user", path="file://${hydra.runtime.cwd}")



import nox  # type: ignore
import os
import requests


@nox.session(name="lint", venv_backend="uv")
@nox.parametrize("python", ["3.10"])
def test_lint(session):
    session.install("-e", ".[dev,onnx-cpu]")
    session.run("ruff", "check", ".")
    session.run("ruff", "format", "--check", ".")
    session.run("ty", "check", ".")


@nox.session(name="test-python", venv_backend="uv")
@nox.parametrize("python", ["3.10", "3.11", "3.12", "3.13"])
@nox.parametrize("torch", ["2.8.0"])
def test_python(session, torch):
    session.install(f"torch=={torch}")
    session.install("-e", ".[dev,onnx-cpu]")
    session.run("pytest")


@nox.session(name="test-torch", venv_backend="uv")
@nox.parametrize("python", ["3.11"])
@nox.parametrize("torch", ["2.4", "2.6", "2.8.0"])
def test_torch(session, torch):
    session.install(f"torch=={torch}")
    session.install("-e", ".[dev,onnx-cpu]")
    session.run("pytest")


@nox.session(name="test-examples", venv_backend="uv", default=False)
@nox.parametrize("python", ["3.11"])
@nox.parametrize("torch", ["2.8.0"])
def test_examples(session, torch):
    session.install(f"torch=={torch}")
    session.install("transformers>=4.0.0")
    session.install("-e", ".[dev,onnx-gpu]")
    files_to_download = [
        ("https://models.silero.ai/models/en/en_v5.onnx", "~/Downloads/silero_model_stt_v5.onnx"),
        (
            "https://github.com/onnx/models/raw/refs/heads/main/validated/vision/classification/resnet/model/resnet50-v2-7.onnx",
            "~/Downloads/resnet50-v2-7.onnx",
        ),
    ]
    print("Downloading example models...")
    for url, path in files_to_download:
        # session.run("wget", url, "-O", os.path.expanduser(path))
        if os.path.isfile(os.path.expanduser(path)):
            print(f"File {path} already exists, skipping download.")
            continue
        print(f"Downloading {url} to {path}...")
        os.path.name
        with open(os.path.expanduser(path), "wb") as f:
            f.write(requests.get(url).content)

    configs = []
    for file in os.listdir(os.path.join("src", "nvbenjo", "conf")):
        file, ext = os.path.splitext(file)
        if ext in [".yml", ".yaml"]:
            configs.append(file)

    for config in configs:
        print(f"Running example {config}:")
        session.run("nvbenjo", "-cn", config)



import logging
from logging import NullHandler
from rich.console import Console


console = Console()

from .benchmark import benchmark_model  # noqa: E402
from .system_info import get_cpu_info, get_gpu_info, get_system_info  # noqa: E402

__all__ = [get_system_info, get_gpu_info, get_cpu_info, benchmark_model]


logging.getLogger(__name__).addHandler(NullHandler())



import os
from os.path import join
from typing import List

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

from . import console
from .utils import format_num, format_seconds


def visualize_results(
    results: pd.DataFrame,
    output_dir: str,
    keys: List[str] = [
        "time_cpu_to_device",
        "time_device_to_cpu",
        "time_inference",
        "time_total_batch_normalized",
        "memory_bytes",
    ],
    hue="runtime_options",
    col="batch_size",
    kind="bar",
):
    sns.set_style("whitegrid")
    for model in results.model.unique():
        mult_devices = len(results.device.unique()) > 1
        for device in results.device.unique():
            model_device_results = results[(results.model == model) & (results.device == device)]
            if len(model_device_results) == 0:
                continue
            for key in keys:
                if key in model_device_results.columns and not model_device_results[key].isnull().all():
                    sns.catplot(
                        data=model_device_results,
                        x="model",
                        y=key,
                        hue=hue,
                        col=col,
                        kind=kind,
                        palette="dark",
                        alpha=0.6,
                    )
                    device_stem = f"{device}_" if mult_devices else ""
                    os.makedirs(join(output_dir, model), exist_ok=True)
                    plt.savefig(join(output_dir, model, f"{device_stem}{key}.png"))
                    plt.close()

    if len(results.device.unique()) == 1 and len(results.model.unique()) > 1:
        for key in keys:
            if key in results.columns and not results[key].isnull().all():
                sns.catplot(
                    data=results.drop_duplicates().reset_index(drop=True),
                    y=key,
                    hue=hue,
                    col=col,
                    kind=kind,
                    row="model",
                    sharey=True,
                    palette="dark",
                    alpha=0.6,
                )
                device_stem = f"{device}_" if mult_devices else ""
                os.makedirs(join(output_dir, "summary"), exist_ok=True)
                plt.savefig(join(output_dir, "summary", f"summary_{key}.png"))
                plt.close()

    # TODO: maybe also check if only one model type then do same for device


def print_system_info(system_info: dict):
    text_color = "white"
    os_info = system_info["os"]
    os_string = os_info["system"].replace("Linux", "Linux 🐧")
    cpu_info = system_info["cpu"]
    gpu_infos = system_info["gpus"]
    driver_version = set(gpu_info["driver"] for gpu_info in gpu_infos)
    driver_version = driver_version.pop() if len(driver_version) == 1 else driver_version

    title = Text("System Information", style="bold cyan")

    content = Text()
    content.append("\n")
    content.append(f"💻️ {os_info['node']}\n", style="bold")
    content.append("OS:   ", style="bold yellow")
    content.append(f"{os_string} - {os_info['version']} ({os_info['release']})\n", style=text_color)
    content.append("CPU:  ", style="bold magenta")
    content.append(f"{cpu_info['model']} ({cpu_info['architecture']})\t", style=text_color)
    content.append("Cores: ", style=f"{text_color} bold")
    content.append(f"{cpu_info['cores']}\n", style=text_color)

    content.append("GPUs", style="bold green")
    if len(gpu_infos) > 0:
        content.append(f" (Driver {driver_version})\n", style="green")
        for gpu_info in gpu_infos:
            content.append("   ", style="bold blue")
            content.append(f"{gpu_info['name']} @ {gpu_info['clock_gpu']} ", style=text_color)
            content.append(f"({gpu_info['memory']} @ {gpu_info['clock_mem']})", style=text_color)
            content.append(f" - {gpu_info['architecture']}\n", style=text_color)
    else:
        content.append("  None\n", style=text_color)

    console.print(Panel(content, title=title, border_style="blue", padding=(1, 4)))


def print_results(results: pd.DataFrame, custom_metric_keys: List[str] = []):
    console.print("\n")
    for model in results.model.unique():
        model_results = results[results.model == model]
        for device in model_results.device.unique():
            # Create a rich table for each model+device combination
            table = Table(
                title=f"Model: {model} on Device: {device}",
                show_header=True,
                header_style="bold",
                show_lines=True,
                title_style="bold",
            )

            # Get grouped results
            device_results = model_results[model_results.device == device]
            device_results = device_results.drop(columns=["device"])
            device_results = device_results.groupby(["model", "runtime_options", "batch_size"]).mean()
            device_results["device"] = device  # Add device column back for display
            print_result = device_results.reset_index()

            # Remove columns where all values are None
            print_result = print_result.dropna(axis="columns", how="all")

            # Format values for display
            for column in print_result.columns:
                if column == "time_total_batch_normalized":
                    top3 = print_result.time_total_batch_normalized.nsmallest(3).index
                    print_result[column] = print_result[column].apply(format_seconds)
                    for i, emoji in enumerate(["🥇", "🥈", "🥉"][: len(top3)]):
                        print_result.loc[top3[i], column] = f"{emoji} {print_result.loc[top3[i], column]}"
                elif column.startswith("time"):
                    print_result[column] = print_result[column].apply(format_seconds)
                elif "bytes" in column:
                    print_result[column] = print_result[column].apply(format_num, bytes=True)
                elif column == "device":
                    print_result[column] = print_result[column].apply(lambda x: f"{x}")
                elif column in custom_metric_keys:
                    print_result[column] = print_result[column].apply(lambda x: f"{x:.3f}")

            # Add columns to the table
            for col in print_result.columns:
                # Set column styles based on data type
                if col == "model":
                    style = "bold green"
                elif col == "runtime_options":
                    style = "bold blue"
                elif col == "batch_size":
                    style = "bold yellow"
                elif col == "time_total_batch_normalized":
                    style = "bold cyan"
                elif col.startswith("time"):
                    style = None
                elif "memory" in col:
                    style = "red"
                else:
                    style = None

                # Format column names for better display
                display_name = col.replace("_", " ").title()
                table.add_column(display_name, style=style, justify="right")

            # Add rows to the table
            for _, row in print_result.iterrows():
                table.add_row(*[str(value) for value in row.values])

            # Display the table in a panel
            console.print(Panel(table, border_style="dim", padding=(1, 2)))
            console.print("\n")
